/* qDSed arm9_main.c QRT 071510 */

#define VERSION "qDSed V0.4 qland.de"

#include <nds.h>
#include <nds/arm9/console.h>
#include <nds/arm9/trig_lut.h>
#include <stdio.h>
#include <fat.h>
#include <sys/dir.h>
#include "debug.h"
#include <ft2build.h>
#include FT_FREETYPE_H
#include FT_GLYPH_H
#include "qdsed.h"
#include "reboot.h"

#define FT_FLOOR(x)     (((x) & -64) / 64)
#define FT_CEIL(x)      ((((x) + 63) & -64) / 64)
#define FT_FIXED(x)     ((x) * 64)

//#define DEBUG_CONSOLE

/* todo

- scrollbar sliding out of slider bounds

- end of scrollbar and file on short files and dirs
- windows .fon support
  (strikes stored are in faces)
- arm7 code for touch events
  (double tap on emulators?)
- variable tab size
- rework line combine
- reworke floatText function
- keyboard
- search function
- bookmarks
- hex display
- special char display

*/

//------------------------------------------------------------------------------

static void loadPrefs();
static void savePrefs();
static void initVideo();
static void initFonts();
static bool loadCachedFont(char*, qfontType*);             
static bool cacheFont(char*, qfontType*);           
static bool renderFont(qfontType*);   
static bool saveCachedFont(char*, qfontType*);  
static void copyFont(qfontType*, qfontType*);
static void loadFontNum(int32);  
static bool showFixedSizes(char*);
static void setView(int16);
static void cleanUp();
static void loadTextNum(int32);
static bool loadText(char*, int32*, bool);
static void floatText(int32, bool, bool, bool);             // first visual position -1 if no seek, update sx, change inversion, update scrollbar
static void drawChar(uint16*, int16, int16, FT_Bitmap*);    // white background      
static void drawCharColor(uint16*, int16, int16, FT_Bitmap*, uint16);   // colored background
static void drawLine(uint16*, int16, int16, int16, int16, uint16);
static void drawRec(uint16*, int16, int16, int16, int16, uint16, uint16);
static void fillRec(uint16*, int16, int16, int16, int16, uint16);
static void invertRec(uint16*, int16, int16, int16, int16);
static void floodFill(uint16*, int16, int16, uint16);
static void fillFour(int16, int16);
static uint16 getPixel(int16, int16);
static void setPixel(int16, int16, uint16);
static uint16 sysTextWidth(char*);
static void drawSysText(char*, int16, int16, uint16);
static void initControlScreen();
static void wait(uint16);
static void status(char*, ...);
static void tabGlyph(qfontType*);
static void showDir(char*, char*, int16);
static void drawScrollbars(uint16*, bool);
static void drawVerScrollbar(uint16*);
static void drawHorScrollbar(uint16*);
static void drawScrollButton(uint16*);
static void scrollVer(uint16*, bool);
static void scrollHor(uint16*, bool);
static void flipScrollbars();
static void copyScrollbars();
static void setVirtualScreenWidth();

//------------------------------------------------------------------------------

uint8 tbuf[TBUFSIZE];               // text buffer
char fontpath[32];                  // fontpath 
char textpath[32];                  // textpath
char prefname[32];                  // prefs name
char stext[STATXTS][STATXTLEN];     // status text
FT_Library library;                 // freetype library
prefsType prefs;                    // preferences
qfontType *curfont;                 // current display font
textType *text;                     // pointer to text struct array in prefs
buttonType *sbut;                   // pointer to status button
uint16 *front, *back, *cont;        // front, back, control video buffer
int32 fsize;                        // file size
int32 tbfipo, tblapo, tblen;        // text buffer first/last file position, text buffer used len
int32 tsfipo, tslapo;               // text screen first/last visual file position
int32 sx, sy;                       // text screen pixel coordinate inside text buffer
int16 tx, ty;                       // touch coordinates on screen
int32 licm;                         // init line counter memory
int16 view;                         // text, text dir, font dir
int16 curt;                         // current text
int32 vica;                         // visible chars
bool rendered=TRUE;                 // FALSE if font parameters are changed
bool gfxrdy=FALSE;                  // avoid drawing before font & grafics are initialized
bool flip=TRUE;                     // flip top/bottom screen
bool seek=FALSE;                    // seek after loading new part of file
int16 scrw=SCREEN_WIDTH;            // screen width variable
int16 scrh=SCREEN_HEIGHT;           // screen height variable
int16 viscwi;                       // virtual screen width
int16 stanum;                       // status text number

//------------------------------------------------------------------------------

int main(void)
{
touchPosition touch;
int32 sxo=0, syo=0;
int16 dx, dy, txo=0, tyo=0;
int16 i=0, m=-1, pup=0;
uint16 *temp;
bool keyl=FALSE, pendown=FALSE, scroba=FALSE;
void (*p)(buttonType*);
uint32 hb=0;                        // horizontal blank periods

    powerON(POWER_ALL_2D);
    irqInit();
    irqEnable(IRQ_VBLANK);
    
    initVideo();
    
    if(!fatInitDefault())           // fatInit(cache_size, set_as_default);   
        ErrorMsg("failed to initialize libfat");
    
    loadPrefs();
    initFonts();    
    initControlScreen();
    status(VERSION);            
            
    //- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    setView(TEXTFILE_VIEW);                     // must be done before loading text to set curt

    if(loadText(text[curt].name, &text[curt].fivipo, TRUE))
        status("could not open file");   
    else
        status("OK %s", text[curt].name);

    floatText(text[curt].fivipo, TRUE, FALSE, TRUE);    // seek, update sx, do not invert line, update scrollbar

    while(TRUE){
        scanKeys();
        touch = touchReadXY();
    
        if(keysHeld() & KEY_L){                 // swap screens with left shoulder button
            if(!keyl){
                if(view)
                    loadTextNum(-1);            // load previous text if dir view
                else
                    lcdSwap();  
            }
            
            keyl = TRUE;
        }
        else{
            keyl = FALSE;
        }

        swiWaitForVBlank();
        
        if(flip){                               // swap screen buffer 
            BG2_CR ^= BG_BMP_BASE(128 / 16);    
            temp = front;
            front = back;
            back = temp;
            flip = FALSE;
        }        

        if(REG_POWERCNT & POWER_SWAP_LCDS){ // sub/control screen on bottom ------------------------
            if(keysHeld() & KEY_TOUCH){
                tx = touch.px;
                ty = touch.py;
                
                if(pendown){                            // slide
                    if(tx>verScroll.posx && ty<butScroll.posy)
                        scrollVer(cont, TRUE);                   
                    else if(ty>horScroll.posy && tx<butScroll.posx)
                        scrollHor(cont, TRUE);
                }
                else{                                   // pen down
                    if(tx > verScroll.posx){
                        if(ty > butScroll.posy){
                            butScroll.status |= PRESSED;
                            drawScrollButton(cont);
                        }
                        else{
                            scrollVer(cont, FALSE);        
                        }
                    }           
                    else if(ty > horScroll.posy){
                        scrollHor(cont, FALSE);
                    }                    
                    else{                    
                        i = 0;
                        
                        do{
                            if(tx>button[i].posx && tx<button[i].posx+button[i].sizex && ty>button[i].posy && ty<button[i].posy+button[i].sizey){
                                drawButton(&button[i], BUTDOWN);
                                m = i;
                                break;
                            }
                        }while(!(button[i++].status & LASTBUT));
                    }
                }
                
                pendown = TRUE;
            }    
            else{
                if(pendown){                            // pen up
                    if(m > -1){
                        drawButton(&button[m], BUTUP);
                        p = button[m].function;
                        p(&button[m]);  //(*p)();
                        m = -1;
                    }
                
                    if(butScroll.status & PRESSED){     // scroll button released
                        butScroll.status &= ~PRESSED;
                        drawScrollButton(cont);
                        flipScrollbars();
                    }
                }
    
                pendown = FALSE;
            }
        }   
        else{ // main/text screen on bottom --------------------------------------------------------
            if(keysHeld() & KEY_TOUCH){ 
                tx = touch.px;
                ty = touch.py;
                
                if(pendown){                    // pen was already down, slide
                    if(prefs.sbotex && tx>verScroll.posx && ty<butScroll.posy){
                        scrollVer(cont, TRUE);                 
                        scroba = TRUE;
                    }  
                    else if(prefs.sbotex && ty>horScroll.posy && tx<butScroll.posx){
                        scrollHor(cont, TRUE);
                        scroba = TRUE;
                    }
                    else if(!scroba){ 
                        dx = tx - txo;
                        dy = ty - tyo;

                        if(dx || dy){
                            if(abs(dx) > abs(dy)){
                                sx -= dx;
                                
                                if(sx < 0) 
                                    sx = 0;
                                else if(sx > viscwi-scrw) 
                                    sx = viscwi-scrw;
                            }
                            else{
                                sy -= dy;
                                //if(sy < 0) sy = 0;
                            }
        
                            if(sx!=sxo || sy!=syo){
                                floatText(seek ? tsfipo : -1, FALSE, FALSE, TRUE);  // seek, do not update sx, do not invert line, update scrollbar
                                seek = FALSE;
            
                                if(!view && sy!=syo && ((tbfipo && tsfipo-tbfipo<vica) || (tblapo!=fsize-1 && tblapo-tslapo<vica))){
                                    loadText(text[curt].name, &tsfipo, FALSE);        
                                    sy = sy % curfont->vspace;
                                    seek = TRUE;
                                }
                                
                                sxo = sx; syo = sy;
                                flip = TRUE;
                            }
                        }
                    }
                }
                else{                           // pen was up before, pen down
                    if(tx > verScroll.posx){
                        if(ty > butScroll.posy){
                            butScroll.status |= PRESSED;
                            drawScrollButton(cont);
                        }
                        else if(prefs.sbotex){
                            scrollVer(cont, FALSE);        
                        }
                    }           
                    else if(prefs.sbotex && ty>horScroll.posy){
                        scrollHor(cont, FALSE);
                    }                    
                    
                    if(hb > 30){
                        floatText(-1, FALSE, TRUE, TRUE);
                        flip = TRUE;
                        hb = 0; pup = 0;
                    }
                }
                
                txo = tx; tyo = ty;
                pendown = TRUE;
            }
            else{
                if(pendown){                        // pen was down before, pen up
                    if(butScroll.status & PRESSED){ // scroll button released
                        butScroll.status &= ~PRESSED;
                        drawScrollButton(cont);
                        flipScrollbars();
                    }

                    if(hb<30 && ++pup==2){
                        if(view == TEXTDIR_VIEW) loadTextNum(licm);
                        else if(view == FONTDIR_VIEW) loadFontNum(licm);
                        hb = 0;
                    }
                }
                
                pendown = FALSE;
                scroba = FALSE;
            }
        }
    
        hb++;
    }
    
    return 0;
}

//------------------------------------------------------------------------------

void loadPrefs()
{
FILE* fptr;
    
    strcpy(fontpath, ROOTDIR);                  // build paths
    strcpy(textpath, ROOTDIR);
    strcpy(prefname, ROOTDIR);
    
    if((fptr=fopen(ROOTDIR SYSFONT, "r"))){     // sys.fon found in root -> desmue emulator
        fclose(fptr);                           // all data in root
    }
    else{                                       // NDS real device
        strcat(fontpath, FONTDIR);              // add dirs
        strcat(textpath, TEXTDIR);
        strcat(prefname, PREFDIR);
    }
    
    strcat(prefname, PREFNAME);                 // add name

    prefs.version = 0;
    fptr = fopen(prefname, "rb");
    
    if(fptr){
        fread(&prefs, sizeof(prefsType), 1, fptr);
        fclose(fptr);
    }
    
    if(prefs.version != PREFVERSION)
        memcpy(&prefs, &defaultPrefs, sizeof(prefsType));

    curt = prefs.curt;
    text = &prefs.text[0];

    button[BUT_FILENUM+prefs.curt-1].status |= CHECKED;
    sprintf(button[BUT_SCRWIDTH].text, "%d", text[curt].viscwi);
    button[BUT_WORDWRAP].status = text[curt].wordwrap ? button[BUT_WORDWRAP].status | CHECKED : button[BUT_WORDWRAP].status & ~CHECKED;
    button[BUT_LINECOMB].status = text[curt].linecomb ? button[BUT_LINECOMB].status | CHECKED : button[BUT_LINECOMB].status & ~CHECKED;
    button[BUT_FORCEMONO].status = text[curt].forcemono ? button[BUT_FORCEMONO].status | CHECKED : button[BUT_FORCEMONO].status & ~CHECKED;
    setVirtualScreenWidth();
}

//------------------------------------------------------------------------------

void savePrefs()
{
FILE* fptr;

    status("saving prefs ...");
        
    if((fptr=fopen(prefname, "wb"))){
        fwrite(&prefs, sizeof(prefsType), 1, fptr);
        fclose(fptr);
        status("OK");
    }
    else{
        status("could not write prefs");
    }
}

//------------------------------------------------------------------------------

void initVideo()
{    
    videoSetMode(MODE_5_2D | DISPLAY_BG2_ACTIVE);                       // text screen double buffered
    videoSetModeSub(MODE_5_2D | DISPLAY_BG2_ACTIVE);                    // keyboard screen
    vramSetBankA(VRAM_A_MAIN_BG_0x06000000);                            // VRAM A for text screen buffer 0
    vramSetBankB(VRAM_B_MAIN_BG_0x06020000);                            // VRAM B for text screen buffer 1
    vramSetBankC(VRAM_C_SUB_BG_0x06200000);                             // VRAM C for control screen
    
    BG2_CR = BG_BMP16_256x256;                                          // set control register for text screen buffer 0 
    BG2_XDX = 1<<8;
    BG2_XDY = 0;
    BG2_YDX = 0;
    BG2_YDY = 1<<8;
    BG2_CY = 0;
    BG2_CX = 0;
    
    SUB_BG2_CR = BG_BMP16_256x256;                                      // control screen buffer 0 control register
    SUB_BG2_XDX = 1<<8;
    SUB_BG2_XDY = 0;
    SUB_BG2_YDX = 0;
    SUB_BG2_YDY = 1<<8;
    SUB_BG2_CY = 0;
    SUB_BG2_CX = 0;
    
    front = BG_GFX0;
    back = BG_GFX1;    
    cont = BG_GFX_SUB;
    
#ifdef DEBUG_CONSOLE    
    cont = ((uint16*)0x06040000);
#endif
    
    memset(front, 0xff, SCREEN_WIDTH*SCREEN_HEIGHT*2);  // white
    memset(back, 0xff, SCREEN_WIDTH*SCREEN_HEIGHT*2);   // white
    memset(cont, 0xff, SCREEN_WIDTH*SCREEN_HEIGHT*2);   // white

    //uint16 i, color;
    //color = RGB15(31, 31, 31) | BIT(15);                // white
    //for(i=0; i<SCREEN_WIDTH*SCREEN_HEIGHT; i++)
    //    BG_GFX_SUB[i] = color;

#ifdef DEBUG_CONSOLE
    videoSetModeSub(MODE_0_2D | DISPLAY_BG0_ACTIVE);
    vramSetBankC(VRAM_C_SUB_BG);
    SUB_BG0_CR = BG_MAP_BASE(31);
    BG_PALETTE_SUB[255] = RGB15(31, 31, 31);    // font color white
    consoleInitDefault((u16*)SCREEN_BASE_BLOCK_SUB(31), (u16*)CHAR_BASE_BLOCK_SUB(0), 16);    
#endif    
}

//------------------------------------------------------------------------------

void initFonts()
{
    if(loadCachedFont(SYSFONT, &sysFont))                       // load sysfont
        if(cacheFont(SYSFONT, &sysFont))                        // not found or wrong version, try to render + cache sysfont
            ErrorMsg("could not render system font");             
    
    if(loadCachedFont(TXTFONT, &txtFont))                       // load txtfont
        copyFont(&txtFont, &sysFont);                           // copy sysfont if load failed
    
    strcpy(tmpFont.name, txtFont.name);                         // copy loaded txtfont settings to tmpfont
    tmpFont.height = txtFont.height;                      
    tmpFont.render = txtFont.render;
    tmpFont.hospco = txtFont.hospco;
    tmpFont.vespco = txtFont.vespco;
    tmpFont.vspcnat = txtFont.vspcnat;                          
    tmpFont.vspace = txtFont.vspace;                          
    tmpFont.nethei = txtFont.nethei;                     
    
    sprintf(button[BUT_FONTHEIGHT].text, "%d", tmpFont.height); // fill in font related buttons
    sprintf(button[BUT_RENDERMODE].text, "%s", tmpFont.render?(tmpFont.render==ANTIALIASE?"AAN":"AA2"):"MON");
    sprintf(button[BUT_HORSPC].text, "%d", tmpFont.hospco);
    sprintf(button[BUT_VERSPC].text, "%d", tmpFont.vespco);
    
    gfxrdy = TRUE;                                              // grafics & fonts are ready now
}

//------------------------------------------------------------------------------

bool loadCachedFont(char *name, qfontType *font)
{
FILE* fptr;
FT_BitmapGlyph fb;
uint16 i, size;
char pana[256];
uint32 id ;
bool error=TRUE;                                        // assume error
 
    strcpy(pana, fontpath);                             // path 
    strcat(pana, name);                                 // + filename
                                                        
    if((fptr=fopen(pana, "rb"))){                       // open file
        fread(&id, 4, 1, fptr);                         // read qfont id
                                                        
        if(id == QFONTID){                              // id OK
            fread(font, sizeof(qfontType), 1, fptr);    
                                                     
            for(i=0; i<NUMGLYPHS; i++){              
                font->glyph[i] = malloc(sizeof(FT_BitmapGlyphRec));
                fread(font->glyph[i], sizeof(FT_BitmapGlyphRec), 1, fptr);                
                                                      
                fb = (FT_BitmapGlyph)font->glyph[i];  
                size = fb->bitmap.pitch * fb->bitmap.rows;
                fb->bitmap.buffer = (uint8*)malloc(size);
                fread(fb->bitmap.buffer, size, 1, fptr);
            }                                          
            
            error = FALSE;                              // all OK
        }
                
        fclose(fptr);
    }    
    
    return error;                              
}

//------------------------------------------------------------------------------
// dst = name of qDSed cached font, font = target font, (tmpFont = source font)
//------------------------------------------------------------------------------

bool cacheFont(char *dst, qfontType *font)
{
FT_BitmapGlyph bs, bd;                                              // source, destination glyph bitmap
uint16 i, size;
bool rend_error, save_error=FALSE;

    status("rendering font ...");

    if(!(rend_error=renderFont(&tmpFont))){                         // render temp font
        status("caching font ...");                                 // rendering was successful

        if(saveCachedFont(dst, &tmpFont)){                          // cache temp font
            status("could not cache font");
            save_error = TRUE;
        }

        if(gfxrdy){                                                   
            for(i=0; i<NUMGLYPHS; i++){                                
                free(((FT_BitmapGlyph)font->glyph[i])->bitmap.buffer);  // free current bitmap
                free(font->glyph[i]);                                   // and glyph
            }
        }

        memcpy(font, &tmpFont, sizeof(qfontType));                   // copy struct
             
        for(i=0; i<NUMGLYPHS; i++){                             
            font->glyph[i] = malloc(sizeof(FT_BitmapGlyphRec));     // copy glyph from temp
            memcpy(font->glyph[i], tmpFont.glyph[i], sizeof(FT_BitmapGlyphRec));

            bs = (FT_BitmapGlyph)tmpFont.glyph[i];                  // source glyph temp bitmap
            bd = (FT_BitmapGlyph)font->glyph[i];                    // target glyph bitmap
            
            size = bs->bitmap.pitch * bs->bitmap.rows;              // copy glyph bitmap
            bd->bitmap.buffer = (uint8*)malloc(size);
            memcpy(bd->bitmap.buffer, bs->bitmap.buffer, size);
        
            FT_Done_Glyph(tmpFont.glyph[i]);                        // free memory of temp glyph
        }
        
        FT_Done_FreeType(library);                                  // close library
        if(!save_error) status("OK");
    }

    return rend_error;
}

//------------------------------------------------------------------------------

bool renderFont(qfontType *font)
{
FT_Error error; 
FT_Face face;
int16 ascent, vs, ad;
uint16 i, loadflags=0, renderflags=0;
char pana[256];

    strcpy(pana, fontpath);                         // path 
    strcat(pana, font->name);                       // + filename

    if(font->render == MONOCHROME){
        loadflags = FT_LOAD_MONOCHROME | FT_LOAD_NO_AUTOHINT;   // FT_LOAD_TARGET_MONO  FT_LOAD_NO_HINTING
        renderflags = FT_RENDER_MODE_MONO;
    }
    else if(font->render == ANTIALIASE){
        loadflags = FT_LOAD_DEFAULT | FT_LOAD_NO_BITMAP;
        renderflags = FT_RENDER_MODE_NORMAL;
    }
    else if(font->render == ANTIALIASE2){
        loadflags = FT_LOAD_DEFAULT | FT_LOAD_NO_BITMAP | FT_LOAD_NO_AUTOHINT;
        renderflags = FT_RENDER_MODE_NORMAL;                   
    }
    
    error = FT_Init_FreeType(&library);
    
    if(error){
        status("could not open FreeType library");
        return error;
    }

    error = FT_New_Face(library, pana, 0, &face);
    
    if(error == FT_Err_Unknown_File_Format){
        status("unknown font file format");
        return error;
    }
    else if(error){  
        status("cannot open font file");
        return error;
    }
    
    //error = FT_Select_Size(face, 2);
    error = FT_Set_Pixel_Sizes(face, 0, font->height);
    
    if(error){
        status("could not set pixel size");
        return error;
    }

    ascent = FT_CEIL(FT_MulFix(face->bbox.yMax, face->size->metrics.y_scale));
    vs = FT_CEIL(FT_MulFix(face->bbox.yMax, face->size->metrics.y_scale));
    font->vspcnat = vs;
    vs *= (100 + font->vespco);
    font->vspace = (vs + vs % 100) / 100;

    for(i=0; i<NUMGLYPHS; i++){
        FT_Load_Char(face, i, loadflags);                                       // FT_Get_Glyph(face->glyph, &glyph[i]); 
        FT_Get_Glyph(face->glyph, &font->glyph[i]); 

        if(face->glyph->format != FT_GLYPH_FORMAT_BITMAP)
            FT_Glyph_To_Bitmap(&font->glyph[i], renderflags, FALSE, TRUE);      // destroys glyph from FT_Get_Glyph
        
        font->metrics[i].bearX = FT_FLOOR(face->glyph->metrics.horiBearingX);   // copy metrics before next glyph is handled
        font->metrics[i].a_bearY = ascent - FT_FLOOR(face->glyph->metrics.horiBearingY);
        ad = FT_CEIL(face->glyph->metrics.horiAdvance);
        font->metrics[i].advnat = ad;
        ad *= (100 + font->hospco);
        font->metrics[i].advance = (ad + ad % 100) / 100;
    }   

    tabGlyph(font);
    font->nethei = ((FT_BitmapGlyph)font->glyph['|'])->bitmap.rows;
    FT_Done_Face(face);                             // close face
    return FALSE;                                   // all OK
}

//------------------------------------------------------------------------------

bool saveCachedFont(char *name, qfontType *font)
{
FT_BitmapGlyph fb;
FILE *fptr;
uint16 i;
char pana[256];
uint32 id=QFONTID;

    strcpy(pana, fontpath);                         // path 
    strcat(pana, name);                             // + filename
    
    if(!(fptr=fopen(pana, "wb")))                   // open file
        return TRUE;                                // error
    
    fwrite(&id, 4, 1, fptr);                        // write qfont id
    fwrite(font, sizeof(qfontType), 1, fptr);       // write font struct

    for(i=0; i<NUMGLYPHS; i++){                     // write glyph + bitmap
        fb = (FT_BitmapGlyph)font->glyph[i];
        fwrite(fb, sizeof(FT_BitmapGlyphRec), 1, fptr);       
        fwrite(fb->bitmap.buffer, fb->bitmap.pitch*fb->bitmap.rows, 1, fptr);
        // fflush(fptr); 
    }

    fclose(fptr);
    return FALSE;                                   // all OK
}

//------------------------------------------------------------------------------

void copyFont(qfontType *dst, qfontType *src)
{
FT_BitmapGlyph bs, bd;                                          // source, destination glyph bitmap
uint16 i, size;

    memcpy(dst, src, sizeof(qfontType));                        // copy struct

    for(i=0; i<NUMGLYPHS; i++){                             
        dst->glyph[i] = malloc(sizeof(FT_BitmapGlyphRec));      // copy glyph
        memcpy(dst->glyph[i], src->glyph[i], sizeof(FT_BitmapGlyphRec));

        bs = (FT_BitmapGlyph)src->glyph[i];                     // source glyph bitmap
        bd = (FT_BitmapGlyph)dst->glyph[i];                     // target glyph bitmap
        
        size = bs->bitmap.pitch * bs->bitmap.rows;              // copy glyph bitmap
        bd->bitmap.buffer = (uint8*)malloc(size);
        memcpy(bd->bitmap.buffer, bs->bitmap.buffer, size);
    }
}

//------------------------------------------------------------------------------

void loadFontNum(int32 num)
{
char name[256];
int32 i, n;
int16 p;

    setView(TEXTFILE_VIEW);                             // must be done before to set curt
    
    if(num > 0){                                        // -1 = empty space, 0 = headline
        p = 0; n = 0;
        
        for(i=0; i<tblen; i++){
            if(n == num) name[p++] = tbuf[i];
            else if(n > num) break;
            if(tbuf[i] == LF) n++;
        }

        name[p-1] = '\0';
        
        if(!showFixedSizes(name)){                          // show fixed sizes of font->face 0
            strcpy(tmpFont.name, name);                     // copy new fontname
            rendered = FALSE;                               // font changed but not rendered
        }
    }

    if(loadText(text[curt].name, &text[curt].fivipo, TRUE))
        status("could not reopen file");
  
    floatText(text[curt].fivipo, TRUE, FALSE, TRUE);    // seek, update sx, do not invert line, update scrollbar
    flip = TRUE;                                        // force refresh of front screen
    lcdSwap();                                          // control to bottom screen
}

//------------------------------------------------------------------------------

bool showFixedSizes(char *name)
{
FT_Error error; 
FT_Face face;
char pana[256];
int16 i;

    strcpy(pana, fontpath);                         // path 
    strcat(pana, name);                             // + filename
    
    error = FT_Init_FreeType(&library);
    
    if(error){
        status("could not open FreeType library");
        return error;
    }

    error = FT_New_Face(library, pana, 0, &face);   // open face 0 
    
    if(error == FT_Err_Unknown_File_Format){
        status("unknown font file format");
        FT_Done_FreeType(library);                  // close library
        return error;
    }
    else if(error){  
        status("cannot open font file");
        FT_Done_FreeType(library);                  // close library
        return error;
    }
    
    sprintf(pana, "OK %s (%d) ", name, (int32)face->num_glyphs);
    
    if(face->num_fixed_sizes){
        for(i=0; i<face->num_fixed_sizes; i++){
            sprintf(pana+strlen(pana), "%c%d", i?' ':'[', (int16)FT_CEIL(face->available_sizes[i].y_ppem));
            if(strlen(pana) > 248) break; 
        }
        
        strcat(pana, "]");
    }

    FT_Done_Face(face);                     // close face
    FT_Done_FreeType(library);              // close library
    status(pana);                           // show face info
    return FALSE;                           // all OK
}

//------------------------------------------------------------------------------

void setView(int16 v)
{
    if(v == TEXTFILE_VIEW){
        curfont = &txtFont;
        curt = prefs.curt;
    }
    else{
        curfont = &sysFont;
        curt = 0;
    }
    
    setVirtualScreenWidth();    
    view = v; 
}

//------------------------------------------------------------------------------

void cleanUp()
{
uint16 i;

    for(i=0; i<NUMGLYPHS; i++){
        free(((FT_BitmapGlyph)sysFont.glyph[i])->bitmap.buffer);
        free(sysFont.glyph[i]);
        free(((FT_BitmapGlyph)txtFont.glyph[i])->bitmap.buffer);
        free(txtFont.glyph[i]);
    }
}

//------------------------------------------------------------------------------

void loadTextNum(int32 num)
{
char name[256];
int32 i, n;
int16 p;

    setView(TEXTFILE_VIEW);                             // must be done before to set curt
    
    if(num > 0){                                        // -1 = empty space, 0 = headline
        p = 0; n = 0;
        
        for(i=0; i<tblen; i++){
            if(n == num) name[p++] = tbuf[i];
            else if(n > num) break;
            if(tbuf[i] == LF) n++;
        }

        name[p-1] = '\0';
        strcpy(text[curt].name, name);                  // copy new filename
        text[curt].fivipo = 0;                          // reset file position
    }

    if(loadText(text[curt].name, &text[curt].fivipo, TRUE))
        status("could not open file");
    else
        status("OK %s", text[curt].name);

    setVirtualScreenWidth();
    floatText(text[curt].fivipo, TRUE, FALSE, TRUE);    // seek, update sx, do not invert line, update scrollbar
    flip = TRUE;                                        // force refresh of front screen
    lcdSwap();                                          // control to bottom screen
}

//------------------------------------------------------------------------------

bool loadText(char *name, int32 *fivipo, bool getsize)
{
char pana[256];
FILE* fptr;
int32 n=0, len;

    strcpy(pana, textpath);
    strcat(pana, name);
    
    if(!(fptr = fopen(pana, "rb"))){
        tblen = tbfipo = tblapo = fsize = 0;
        return TRUE;        // error, could not open file
    }

    if(getsize){
        fseek(fptr, 0, SEEK_END);
        fsize = ftell(fptr);
    }
        
    if(*fivipo >= fsize)
        *fivipo = 0;        // correct first visual position
    
    if(fsize <= TBUFSIZE){
        tbfipo = 0;
        len = fsize;
    }
    else{    
        if(vica > TBUFSIZE/3)
            status("warning: low buffer");

        if(vica)    // old version: tbfipo = *fivipo - TBUFSIZE/4;                           
            tbfipo = *fivipo - (TBUFSIZE - vica) / 2;       
        else
            tbfipo = *fivipo - (TBUFSIZE - TBUFSIZE/8) / 2;

        len = TBUFSIZE;
        if(tbfipo < 0) tbfipo = 0;
        if(tbfipo+len >= fsize) tbfipo = fsize - len;
    }
    
    fseek(fptr, tbfipo, SEEK_SET);

    while(n < len)
        tbuf[n++] = fgetc(fptr);
    
    tblapo = tbfipo + len - 1;
    tblen = len;    
    fclose(fptr);
    //printf("%d %d %d\n",tbfipo, *fivipo, tblapo);
    return FALSE;       // no error
}

//------------------------------------------------------------------------------

#define SIZEX   (b->sizex-1)
#define SIZEY   (b->sizey-1)
#define POSX    (b->posx)
#define POSY    (b->posy)

void drawButton(buttonType *b, uint16 draw)
{
uint16 color;

    color = b->color1;

    if(draw & BUTDOWN)
        color = INVCOL;
    else
        if(b->status & CHECKED) 
            color = b->color2;
    
    if(b->status & TRILEFT){
        if(draw & INIT){
            drawLine(cont, POSX+SIZEX-1, POSY, POSX+1, POSY+SIZEY/2-1, RGB15(0, 0, 0));
            drawLine(cont, POSX, POSY+SIZEY/2, POSX, POSY+SIZEY/2+1, RGB15(0, 0, 0));
            drawLine(cont, POSX+SIZEX-1, POSY+SIZEY, POSX+1, POSY+SIZEY/2+2, RGB15(0, 0, 0));
            drawLine(cont, POSX+SIZEX, POSY+SIZEY-1, POSX+SIZEX, POSY+1, RGB15(0, 0, 0));
        }
        
        floodFill(cont, POSX + SIZEX/2, POSY+SIZEY/2, color);
    }
    else if(b->status & TRIRIGHT){
        if(draw & INIT){
            drawLine(cont, POSX+1, POSY, POSX+SIZEX-1, POSY+SIZEY/2-1, RGB15(0, 0, 0));
            drawLine(cont, POSX+SIZEX, POSY+SIZEY/2, POSX+SIZEX, POSY+SIZEY/2+1, RGB15(0, 0, 0));
            drawLine(cont, POSX+1, POSY+SIZEY, POSX+SIZEX-1, POSY+SIZEY/2+2, RGB15(0, 0, 0));
            drawLine(cont, POSX, POSY+SIZEY-1, POSX, POSY+1, RGB15(0, 0, 0));
        }
        
        floodFill(cont, POSX + SIZEX/2, POSY+SIZEY/2, color);
    }
    else{
        if(draw & INIT)
            drawRec(cont, POSX, POSY, POSX+SIZEX, POSY+SIZEY, 0x8000, b->status);   // 0x8000, do not fill rec

        fillRec(cont, POSX+1, POSY+1, POSX+SIZEX-1, POSY+SIZEY-1, color);
        
        if(b->status & STATBUT){
            drawSysText(stext[stanum], POSX + 2, POSY + (SIZEY - (sysFont.nethei*17)/10) / 2, color);
            sbut = b;
        }
        else{
            drawSysText(b->text, POSX + (SIZEX - sysTextWidth(b->text)) / 2 + 1, POSY + (SIZEY - (sysFont.nethei*17)/10) / 2, color);
        }
    }
}

//------------------------------------------------------------------------------

void floatText(int32 fivipo, bool updx, bool chinv, bool upscba)
{
static int32 x1, x2, y1, y2, maxsy=0x7fffffff;
FT_BitmapGlyph bglyph;
metricsType *metrics;
int32 i, fx, fy, fip, lap, wmx, j, fxx, li, lic=0;
int16 ad=0, vspace;
uint8 c, d, lc;
bool vis;

    memset(back, 0xff, SCREEN_WIDTH*SCREEN_HEIGHT*2);   // delete background

    vis = fivipo < 0;                   // fivipo==-1 -> no seek -> vis=TRUE
    // if(fivipo == tblapo) vis = TRUE;
    wmx = (viscwi * 3) / 4;             // word wrap min x
    vspace = curfont->vspace;
    metrics = curfont->metrics;
    if(chinv) licm = -1;                // change inversion on tap down, delete line counter mem

    fx = 0; fy = 0;                     // render pixel coordinates in frame
    fip = -1; lap = 0;                  // init text screen first and last position 
    lc = 255; li = 0;                   // init last char, and last i

    if(updx){
        maxsy = 0x7fffffff;             // reset maxsy
        vica = 0;                       // reset visual chars
    }
    
    if(vis)                             // if no seek
        if(sy < 0) sy = 0;              // check for boundaries

    if(tblapo==fsize-1 && sy>maxsy)     // check for end of text
        sy = maxsy;

    for(i=0; i<tblen; i++){
        if(!vis && tbfipo+i==fivipo+1){            
            if(updx){
                if(fx-ad < viscwi-scrw)
                    sx = fx - ad;       // position of first visual char, ad holds last advance
                else
                    sx = viscwi - scrw;
                     
                if(sx < 0) sx = 0;                    
            }
            
            sy += fy;              
            if(sy < 0) sy = 0;

            i = li; fx = 0;             // restore i of last return, render complete line
            vis = TRUE;
        }

        c = tbuf[i];

        if(c==LF || c==CR){
            if(vis && fy>sy-vspace && fx>=sx && fx<sx+scrw){
                if(fip < 0) fip = i;
                lap = i;
            }

            if(c == LF){
                if(text[curt].linecomb && i>2){
                    if(tbuf[i-2]==' ' && tbuf[i-3]>' '){    // CHR SPC CR LF
                        continue;
                    }
                }
            }
            else{                   // CR
                continue;             
            }

            lic++;
            NEWLINE(1)              // LF
            continue;
        }

        ad = metrics[c].advance;
        
        if(text[curt].wordwrap){
            if(fx>wmx && lc<'!' && c>' '){
                j = i + 1;
                fxx = fx + ad;
            
                while(fxx<viscwi && j<tblen && (d=tbuf[j++])>=' '){
                    fxx += metrics[d].advance;
                    if(d == ' ') break;
                }
                
                if(fxx >= viscwi)
                    NEWLINE(0)
            }
        }
        
        if(fx+ad >= viscwi)
            NEWLINE(0)
        
        if(vis && fy>sy-vspace && fx>sx-ad && fx<sx+scrw){
            bglyph = (FT_BitmapGlyph)curfont->glyph[c];
            drawChar(back, fx-sx+metrics[c].bearX, fy-sy+metrics[c].a_bearY, &bglyph->bitmap);
            if(fip < 0) fip = i;
            lap = i;
        
            if(chinv && ty>fy-sy && ty<fy-sy+vspace+2){
                y1 = fy; y2 = y1;
                x1 = fx; x2 = x1;
                licm = lic;
                chinv = FALSE;
            }
        
            if(view && licm==lic){
                if(y2 < fy) y2 = fy;
                if(x2 < fx+ad+1) x2 = fx+ad+1;
            }
        }
        
        lc = c;
        fx += ad;        
    }    

    if(view && !chinv){
        if(tx < x2-sx)
            invertRec(back, x1-sx, y1-sy+2, x2-sx, y2-sy+vspace+2);
        else
            licm = -1;
    }

    if(fip >= 0){
        tsfipo = tbfipo + fip;
        tslapo = tbfipo + lap;
        if(vica < tslapo-tsfipo) vica = tslapo - tsfipo;
    }

    if(tblapo==fsize-1 && maxsy==0x7fffffff && tblen-i==0)
        maxsy = fy - vspace;

    if(upscba)
        drawScrollbars(cont, updx); 
    
    if(prefs.sbotex)
        copyScrollbars();
    
    //printf("%d % d %d\n", (tsfipo-tbfipo)*100/tblen, (tslapo-tsfipo)*100/tblen, (tblapo-tslapo)*100/tblen);
    //printf("%d %d %d\n", tsfipo, tslapo, tslapo-tsfipo);
}

//------------------------------------------------------------------------------

void drawChar(uint16 *dbuf, int16 x, int16 y, FT_Bitmap *bitmap)
{
uint16 *dest = dbuf + SCREEN_WIDTH * y + x;
uint8 *src = bitmap->buffer;
int16 width = bitmap->width;
int16 height = bitmap->rows;
int16 wa, wb, wc, n, xx, yy;
uint8 g, *s;

    yy = y;

    if(bitmap->pixel_mode == FT_PIXEL_MODE_GRAY){          
        while(height--){
            s = src;
            wa = width;
            xx = x;
            
            while(wa--){
                g = *(s++);
    
                if(g && xx>-1 && xx<scrw && yy>-1){
                    g = (0xff-g)>>3;
                    *dest = RGB15(g, g, g) | BIT(15);
                }
                
                xx++;
                dest++;
            }
    
            yy++;
            src += bitmap->pitch;
            dest += SCREEN_WIDTH - width;
        }
    }
    else if(bitmap->pixel_mode == FT_PIXEL_MODE_MONO){
        wa = ((width - 1) >> 3) + 1;        // num bytes in width
        wc = width % 8 ? width % 8 : 8;     // num relic bits in width
        
        while(height--){
            wb = wa;                        // copy num bytes in width  
            s = src;                        // copy source pointer
            xx = x;                         // copy x coord

            while(wb--){                    // count num bytes in width down
                g = *(s++);                 // get source byte
                n = wb ? 8 : wc;            // num bits to go
                
                while(n--){                 // count down bits
                    if(g&0x80 && xx>-1 && xx<scrw && yy>-1) 
                        *dest = BIT(15);    // set dest word
                    
                    g <<= 1;                // next bit
                    xx++;                   // advance x coord
                    dest++;                 // advance dest pointer
                }
            }

            yy++;                           // advance y coord
            src += bitmap->pitch;           // advance source pointer
            dest += SCREEN_WIDTH - width;   // advance dest pointer
        }
    }
}

//------------------------------------------------------------------------------

void drawCharColor(uint16 *dbuf, int16 x, int16 y, FT_Bitmap *bitmap, uint16 color)
{
uint16 *dest = dbuf + SCREEN_WIDTH * y + x;
uint8 *src = bitmap->buffer;
int16 width = bitmap->width;
int16 height = bitmap->rows;
int16 wa, wb, wc, n, xx, yy;
uint8 g, *s;
uint8 rr, gg, bb;

    yy = y;
    rr = color & 0x001f;
    gg = (color >> 5) & 0x001f;
    bb = color >> 10;

    if(bitmap->pixel_mode == FT_PIXEL_MODE_GRAY){          
        while(height--){
            s = src;
            wa = width;
            xx = x;
            
            while(wa--){
                g = *(s++);
    
                if(g && xx>-1 && xx<scrw && yy>-1){
                    g = 0xff - g;
                    *dest = RGB15(g*rr>>8, g*gg>>8, g*bb>>8) | BIT(15);
                }
                
                xx++;
                dest++;
            }
    
            yy++;
            src += bitmap->pitch;
            dest += SCREEN_WIDTH - width;
        }
    }
    else if(bitmap->pixel_mode == FT_PIXEL_MODE_MONO){
        wa = ((width - 1) >> 3) + 1;
        wc = width % 8 ? width % 8 : 8;
        
        while(height--){
            wb = wa;      
            s = src;
            xx = x;

            while(wb--){            
                g = *(s++);
                n = wb ? 8 : wc;
                
                while(n--){
                    if(g&0x80 && xx>-1 && xx<scrw && yy>-1) 
                        *dest = BIT(15);
                    
                    g <<= 1;
                    xx++;
                    dest++;
                }
            }

            yy++;
            src += bitmap->pitch;
            dest += SCREEN_WIDTH - width;
        }
    }
}

//------------------------------------------------------------------------------

void drawLine(uint16 *dbuf, int16 x1, int16 y1, int16 x2, int16 y2, uint16 color)
{
int16 yStep = SCREEN_WIDTH;
int16 xStep = 1;      
int16 xDiff = x2 - x1;
int16 yDiff = y2 - y1;
int16 errorTerm = 0;
int32 offset = y1 * SCREEN_WIDTH + x1; 
int16 i; 
    
    color |= BIT(15);
    
    if(yDiff < 0){                      // need to adjust if y1 > y2            
       yDiff = -yDiff;                  // absolute value
       yStep = -yStep;                  // step up instead of down   
    }
    
    if(xDiff < 0){                      // same for x
       xDiff = -xDiff;            
       xStep = -xStep;            
    }        
    
    xDiff++; yDiff++;
    
    if(xDiff > yDiff){                  // case for changes more in X than in Y       
       for(i=0; i<xDiff; i++){                           
          dbuf[offset] = color;  
          offset += xStep;           
          errorTerm += yDiff;     
 
          if(errorTerm >= xDiff){  
             errorTerm -= xDiff;     
             offset += yStep;        
          }
       }
    }
    else{                               // case for changes more in Y than in X
       for(i=0; i<yDiff; i++){  
          dbuf[offset] = color;  
          offset += yStep;           
          errorTerm += xDiff;    
 
          if(errorTerm >= yDiff){     
             errorTerm -= yDiff;  
             offset += xStep;     
          }
       }
    }
}

//------------------------------------------------------------------------------

void drawRec(uint16 *buf, int16 x1, int16 y1, int16 x2, int16 y2, uint16 color, uint16 flags)
{
    if(flags & LINETOP)
        drawLine(buf, x1+1, y1, x2-1, y1, RGB15(0, 0, 0));

    if(flags & LINERIGHT)
        drawLine(buf, x2, y1+1, x2, y2-1, RGB15(0, 0, 0));

    if(flags & LINEBOTTOM)
        drawLine(buf, x1+1, y2, x2-1, y2, RGB15(0, 0, 0));

    if(flags & LINELEFT)
        drawLine(buf, x1, y1+1, x1, y2-1, RGB15(0, 0, 0));

    if(!(color&0x8000))
        fillRec(buf, x1+1, y1+1, x2-1, y2-1, color);
}

//------------------------------------------------------------------------------

void fillRec(uint16 *dbuf, int16 x1, int16 y1, int16 x2, int16 y2, uint16 color)
{
int16 xDiff = x2 - x1;
int16 yDiff = y2 - y1;    
int16 xStep = 1;
int16 yStep;
int32 offset = y1 * SCREEN_WIDTH + x1; 
int16 i, j; 

    color |= BIT(15);
   
    if(xDiff < 0){                     
       xDiff = -xDiff;            
       xStep = -xStep;            
    }        

    yStep = SCREEN_WIDTH - (xDiff + 1); 

    if(yDiff < 0){                            
       yDiff = -yDiff;                 
       yStep = -yStep;
    }
   
    xDiff++; yDiff++;
   
    for(j=0; j<yDiff; j++){
        for(i=0; i<xDiff; i++){                           
            dbuf[offset] = color;  
            offset += xStep;
        }
        
        offset += yStep;        
    }
}

//------------------------------------------------------------------------------

void invertRec(uint16 *dbuf, int16 x1, int16 y1, int16 x2, int16 y2)
{
int16 xDiff;
int16 yDiff;    
int16 xStep = 1;
int16 yStep;
int32 offset; 
int16 i, j; 
   
    if(x1 < 0) x1 = 0;
    if(x2 >= SCREEN_WIDTH) x2 = SCREEN_WIDTH-1;
    if(y1 < 0) y1 = 0;

    xDiff = x2 - x1;
    yDiff = y2 - y1;    
    offset = y1 * SCREEN_WIDTH + x1; 
   
    if(xDiff < 0){                     
       xDiff = -xDiff;            
       xStep = -xStep;            
    }        

    yStep = SCREEN_WIDTH - (xDiff + 1); 

    if(yDiff < 0){                            
       yDiff = -yDiff;                 
       yStep = -yStep;
    }
   
    for(j=0; j<yDiff+1; j++){
        for(i=0; i<xDiff+1; i++){                           
            dbuf[offset] ^= 0x7fff;  
            offset += xStep;
        }
        
        offset += yStep;        
    }
}

//------------------------------------------------------------------------------

uint16 fill_color, seed_color, *scrbuf;

void floodFill(uint16 *buf, int16 x, int16 y, uint16 color)
{
    scrbuf = buf;
    fill_color = color;
    seed_color = getPixel(x, y);

    if(fill_color != seed_color)
        fillFour(x, y);
}

//------------------------------------------------------------------------------

void fillFour(int16 x, int16 y)
{
    if(seed_color == getPixel(x, y)){ 
        setPixel(x, y, fill_color);
        fillFour(x+1, y);
        fillFour(x-1, y);
        fillFour(x, y+1);
        fillFour(x, y-1);
    }
}

//------------------------------------------------------------------------------

uint16 getPixel(int16 x, int16 y)
{
int32 offset = y * SCREEN_WIDTH + x;

    return(scrbuf[offset] & 0x7fff);
}

//------------------------------------------------------------------------------

void setPixel(int16 x, int16 y, uint16 fill_color)
{
int32 offset = y * SCREEN_WIDTH + x;

    scrbuf[offset] = fill_color | BIT(15);
}

//------------------------------------------------------------------------------

uint16 sysTextWidth(char *text)
{
uint16 w=0;
char *p=text; 
uint8 c;

    while((c=*p++))
        w += sysFont.metrics[c].advance;

    return w;
}

//------------------------------------------------------------------------------

void drawSysText(char *text, int16 x, int16 y, uint16 color)
{
char *p=text;
uint8 c;

    while((c=*p++)){
        drawCharColor(cont, x+sysFont.metrics[c].bearX, y+sysFont.metrics[c].a_bearY, &((FT_BitmapGlyph)sysFont.glyph[c])->bitmap, color);
        x += sysFont.metrics[c].advance;
    }
}

//------------------------------------------------------------------------------

void initControlScreen()
{
int16 i=0;
    
    do{
        drawButton(&button[i], INIT);
    }while(!(button[i++].status & LASTBUT));
}

//------------------------------------------------------------------------------

void wait(uint16 t)
{
    TIMER_CR(0) = 0;                                // stop timer 0
    TIMER_DATA(0) = -3277;                          // set timer 0 to 0.1 s / ((2^25 Hz)^-1 * 1024
    TIMER_CR(0) = TIMER_DIV_1024 | TIMER_ENABLE;    // start timer 0
    
    TIMER_CR(1) = 0;                                // stop timer 1
    TIMER_DATA(1) = 0;                              // reset timer 1
    TIMER_CR(1) = TIMER_CASCADE | TIMER_ENABLE;     // cascade and start timer 1
    
    while(getTimer(1) < t)                          // wait t * 0.1 s
        ;
}

//------------------------------------------------------------------------------

static void status(char *format, ...)
{
va_list argp;
char text[256], *p, *pp;
int16 n, a;

    if(gfxrdy){
        va_start(argp, format);
        vsprintf(text, format, argp);
        va_end(argp);
        
        p = text;
        n = 0;
        
        while(n<STATXTS && p<text+strlen(text)){    // cut status string into pieces
            a = STATXTLEN - 4;
            
            if(strlen(p) > a){
                pp = p + a - 1;
                
                while(*pp!=' ' && pp>p+a-5)
                    pp--;

                if(*pp == ' ') a = pp-p+1;
            }
            else{
                a = strlen(p);
            }
            
            strncpy(stext[n], p, a);
            stext[n][a] ='\0';
            p += a;
            
            if(p < text+strlen(text))
                strcat(stext[n], "...");
            
            n++;
        }
        
        if(n < STATXTS)
            strcpy(stext[n], VERSION);
        
        for(a=n+1; a<STATXTS; a++)
            strcpy(stext[a], "-");
        
        stanum = 0;        
        drawButton(sbut, REFRESH);    
    }
}

//------------------------------------------------------------------------------

static void tabGlyph(qfontType *font)
{
    font->metrics[TAB].advance = TABSIZE * font->metrics[' '].advance;
    font->metrics[TAB].advnat = TABSIZE * font->metrics[' '].advnat;
    ((FT_BitmapGlyph)font->glyph[TAB])->bitmap.width = 0;
    ((FT_BitmapGlyph)font->glyph[TAB])->bitmap.rows = 0;
}

//------------------------------------------------------------------------------

void showDir(char *pana, char *kind, int16 view)
{
struct stat st;
char name[256];
DIR_ITER* dir;
char *buf;

    status("reading dir ...");
    text[curt].fivipo = tsfipo;                     // save current text position
    dir = diropen(pana);
    
    if(dir == NULL){
        status("%s directory not found", kind);
        return;
    }
    else{
        buf = (char*)tbuf;
        sprintf(buf, "Load %c%s (exit):\n", *kind&0xdf, kind+1);    // first letter upper case
        tblen = strlen(buf);

        while(tblen<TBUFSIZE-257 && dirnext(dir, name, &st)==0){
            if(st.st_mode & S_IFDIR){   // directory
            }
            else{                       // file
                strcat(buf, name);
                strcat(buf, "\n");
                tblen += strlen(name) + 1;
            }
        }

        // dirreset(dir);
        dirclose(dir);
    }
       
    fsize = tblen;                      // needed for scrollbars
    tbfipo = 0;
    tblapo = fsize-1;
    
    sx = 0; sy = 0;                     // init text screen position
    tx = 1; ty = 1;                     // imitate tap into headline
    setView(view);                      // set view
    floatText(0, TRUE, TRUE, TRUE);     // seek, update sx, invert headline, update scrollbar
    flip = TRUE;                        // force refresh of front screen
    status("OK");
    lcdSwap();                          // directory to bottom screen
}

//------------------------------------------------------------------------------

void drawScrollbars(uint16 *buf, bool upversize)
{
static int16 verpos, horpos, horsize;
static bool sbb=FALSE;

    if(tslapo < fsize){     
        if(upversize){
            verScroll.slsize = (tslapo - tsfipo) * verScroll.sizey / fsize;
            if(verScroll.slsize < 10) verScroll.slsize = 10;
        }
        
        verScroll.slpos = tsfipo * (verScroll.sizey-verScroll.slsize) / fsize;
    }
    else{
        if(upversize) verScroll.slsize = verScroll.sizey-1;
        verScroll.slpos = 0;        
    }

    if(upversize || verpos-verScroll.slpos) 
        drawVerScrollbar(buf);
    
    verpos = verScroll.slpos;
    
    //----------------------------------
    
    if(viscwi != scrw){
        horScroll.slsize = scrw * horScroll.sizex / viscwi;
        horScroll.slpos =  sx * horScroll.sizex / viscwi;
    }
    else{
        horScroll.slsize = horScroll.sizex-1;
        horScroll.slpos = 0;
    }

    if(horpos-horScroll.slpos || horsize-horScroll.slsize)
        drawHorScrollbar(buf);

    horpos = horScroll.slpos;
    horsize = horScroll.slsize;

    //----------------------------------

    if(!sbb){   // draw button only once
        drawScrollButton(buf);
        sbb = TRUE;
    }
}

//------------------------------------------------------------------------------

void drawVerScrollbar(uint16 *buf)
{     
scrollbarType *b;
int16 slsize, slpos;
    
    b = &verScroll;
    slsize = b->slsize;
    slpos = b->slpos;

    if(slpos > 1)                                                                               // draw upper part
        drawRec(buf, POSX, POSY, POSX+SIZEX, slpos, b->colbar1, LINETOP|LINERIGHT|LINELEFT);

    if(slpos+slsize < SIZEY-1)                                                                  // draw lower part
        drawRec(buf, POSX, slpos+slsize, POSX+SIZEX, POSY+SIZEY, b->colbar1, LINEBOTTOM|LINERIGHT|LINELEFT);
    
    drawRec(buf, POSX, slpos, POSX+SIZEX, slpos+slsize, b->colsli1, LINEALL);                   // draw slider
    scrbuf = buf;
    setPixel(POSX, slpos, b->colbar1);
    setPixel(POSX+SIZEX, slpos, b->colbar1);
    setPixel(POSX, slpos+slsize, b->colbar1);
    setPixel(POSX+SIZEX, slpos+slsize, b->colbar1);
    
    drawLine(buf, POSX+2, slpos+slsize/2-2, POSX+SIZEX-2, slpos+slsize/2-2, RGB15(0, 0, 0));    // draw slider grip
    drawLine(buf, POSX+2, slpos+slsize/2, POSX+SIZEX-2, slpos+slsize/2, RGB15(0, 0, 0));
    drawLine(buf, POSX+2, slpos+slsize/2+2, POSX+SIZEX-2, slpos+slsize/2+2, RGB15(0, 0, 0));
}

//------------------------------------------------------------------------------

void drawHorScrollbar(uint16 *buf)
{     
scrollbarType *b;    
int16 slsize, slpos;
    
    b = &horScroll;
    slsize = b->slsize;
    slpos = b->slpos;    
    
    if(slpos > 1)                                                                               // draw left part
        drawRec(buf, POSX, POSY, b->slpos, POSY+SIZEY, b->colbar1, LINELEFT|LINETOP|LINEBOTTOM);
    
    if(slpos+slsize < SIZEX-1)                                                                  // draw right part
        drawRec(buf, slpos+slsize, POSY, POSX+SIZEX, POSY+SIZEY, b->colbar1, LINETOP|LINEBOTTOM|LINERIGHT);

    drawRec(buf, slpos, POSY, slpos+b->slsize, POSY+SIZEY, b->colsli1, LINEALL);                // draw slider
    scrbuf = buf;
    setPixel(slpos, POSY, b->colbar1);
    setPixel(slpos, POSY+SIZEY, b->colbar1);
    setPixel(slpos+slsize, POSY, b->colbar1);
    setPixel(slpos+slsize, POSY+SIZEY, b->colbar1);
    
    drawLine(buf, slpos+slsize/2-2, POSY+2, slpos+slsize/2-2, POSY+SIZEY-2, RGB15(0, 0, 0));    // draw slider grip
    drawLine(buf, slpos+slsize/2, POSY+2, slpos+slsize/2, POSY+SIZEY-2, RGB15(0, 0, 0));
    drawLine(buf, slpos+slsize/2+2, POSY+2, slpos+slsize/2+2, POSY+SIZEY-2, RGB15(0, 0, 0));
}

//------------------------------------------------------------------------------

void drawScrollButton(uint16 *buf)
{
buttonType *b;
    
    b = &butScroll;
    drawRec(buf, POSX, POSY, POSX+SIZEX, POSY+SIZEY, butScroll.status&PRESSED?INVCOL:b->color1, b->status);
    scrbuf = buf;
    setPixel(POSX, POSY, b->color2);
    setPixel(POSX+SIZEX, POSY, b->color2);
    setPixel(POSX, POSY+SIZEY, b->color2);
    setPixel(POSX+SIZEX, POSY+SIZEY, b->color2);
}

//------------------------------------------------------------------------------

#define MAXSLPOS    (0x7fffffff / SCREEN_HEIGHT)

void scrollVer(uint16 *buf, bool slide)
{
static int16 oty;
static bool grip=FALSE;
int32 fpos;
int16 dty=0, oslp;
bool sboc;

    if(slide){                                  // slide
        if(grip && (dty=ty-oty)){
            oslp = verScroll.slpos;             // old slider position
            verScroll.slpos += dty;
            if(verScroll.slpos < 0) verScroll.slpos = 0;
            if(verScroll.slpos > verScroll.sizey-verScroll.slsize-1) verScroll.slpos = verScroll.sizey-verScroll.slsize-1;
            
            if(verScroll.slpos != oslp){
                drawVerScrollbar(buf);

                if(fsize>100000 && verScroll.slpos==verScroll.sizey-verScroll.slsize-1){
                    fpos = fsize - vica/3;
                }
                else{
                    if(fsize < MAXSLPOS)
                        fpos = fsize * verScroll.slpos / (verScroll.sizey-verScroll.slsize);            
                    else
                        fpos = (fsize / (verScroll.sizey-verScroll.slsize)) * verScroll.slpos;           
                }
    
                if(!view && ((tbfipo && fpos<tbfipo+vica) || (tblapo!=fsize-1 && fpos>tblapo-vica)))
                    loadText(text[curt].name, &fpos, FALSE);        
                
                sy = 0;            
                floatText(fpos, FALSE, FALSE, FALSE);   // seek, no update sx, no invert line, no update scrollbar
                flip = TRUE;
            }

            dty = 0;
        }
    }
    else{                                       // pen down
        if(ty > verScroll.slpos ){
            if(ty < verScroll.slpos+verScroll.slsize){
                grip = TRUE;                    // ty - verScroll.slpos;
            }
            else{                               // page down
                dty = scrh - curfont->vspace - (sy+scrh)%curfont->vspace;
                grip = FALSE;
            }
        }
        else{                                   // page up
            dty = -(scrh - 2*curfont->vspace + (sy-scrh)%curfont->vspace);
            grip = FALSE;
        }
    }
        
    if(dty){
        sy += dty;

        if(seek && dty<0){                                      // floatText can't seek and render backwards 
            sboc = prefs.sbotex;                                // copy variable
            prefs.sbotex = FALSE;                               // block scrollbar copy
            floatText(tsfipo, FALSE, FALSE, FALSE);             // seek, no update sx, no invert line, no update scrollbar
            prefs.sbotex = sboc;                                // restore variable
            floatText(-1, FALSE, FALSE, TRUE);                  // seek, no update sx, no invert line, update scrollbar
            seek = FALSE;                                       // seek done
        }

        floatText(seek ? tsfipo : -1, FALSE, FALSE, TRUE);      // seek, no update sx, no invert line, update scrollbar
        flip = TRUE;
        seek = FALSE;

        if(!view && ((tbfipo && tsfipo-tbfipo<vica) || (tblapo!=fsize-1 && tblapo-tslapo<vica))){
            loadText(text[curt].name, &tsfipo, FALSE);        
            sy = sy % curfont->vspace;
            seek = TRUE;
        }
    }
        
    oty = ty;
}

//------------------------------------------------------------------------------

void scrollHor(uint16 *buf, bool slide)
{
static int16 otx;
static bool grip=FALSE;
int16 dtx=0, oslp;

    if(slide){                                  // slide
        if(grip)
            dtx = tx - otx;
    }
    else{                                       // pen down
        if(tx > horScroll.slpos ){
            if(tx < horScroll.slpos+horScroll.slsize){
                grip = TRUE;                    // ty - verScroll.slpos;
            }
            else{                               // page right
                dtx = horScroll.slsize - 1;
                grip = FALSE;
            }
        }
        else{                                   // page left
            dtx = -(horScroll.slsize - 1);
            grip = FALSE;
        }
    }
    
    if(dtx){
        oslp = horScroll.slpos;                 // old slider position
        horScroll.slpos += dtx;
        if(horScroll.slpos < 0) horScroll.slpos = 0;
        if(horScroll.slpos > horScroll.sizex-horScroll.slsize-1) horScroll.slpos = horScroll.sizex-horScroll.slsize-1;
        
        if(horScroll.slpos != oslp){
            drawHorScrollbar(buf);
            
            if(horScroll.slpos == horScroll.sizex-horScroll.slsize-1)
                sx = viscwi - scrw;
            else                    
                sx = viscwi * horScroll.slpos / horScroll.sizex;
            
            floatText(-1, FALSE, FALSE, FALSE); // no seek, no update sx, no invert line, no update scrollbar
            flip = TRUE;
        }    
    }    
        
    otx = tx;
}

//------------------------------------------------------------------------------

void flipScrollbars()
{
    if(prefs.sbotex){
        prefs.sbotex = FALSE;
        scrw = SCREEN_WIDTH;
        scrh = SCREEN_HEIGHT;
    }
    else{
        prefs.sbotex = TRUE;
        scrw = SCREEN_WIDTH - verScroll.sizex;
        scrh = SCREEN_HEIGHT - horScroll.sizey;
    }
    
    setVirtualScreenWidth();
    floatText(-1, FALSE, FALSE, TRUE);
    flip = TRUE;
}

//------------------------------------------------------------------------------
// assumes starting points of scrollbars = 0

void copyScrollbars()
{
uint16 y, sizex, sizey;
uint8 *src, *dst;
    
    if(viscwi == SCREEN_WIDTH-verScroll.sizex){                     // vertical bar only
        sizey = SCREEN_HEIGHT * 2;                                  // vertical bar + button
    }
    else{                                                           // vertical + horizontal bar
        src = (uint8*)cont + horScroll.posy * SCREEN_WIDTH * 2;     
        dst = (uint8*)back + horScroll.posy * SCREEN_WIDTH * 2;
        memcpy(dst, src, horScroll.sizey * SCREEN_WIDTH * 2);       // horizontal bar + button

        sizey = verScroll.sizey;                                    // vertical bar
    }

    src = (uint8*)cont + verScroll.posx * 2;                        
    dst = (uint8*)back + verScroll.posx * 2;
    sizex = verScroll.sizex * 2;

    for(y=0; y<sizey; y++){                                         // vertical bar (+ button)
        memcpy(dst, src, sizex);
        dst += SCREEN_WIDTH * 2;
        src += SCREEN_WIDTH * 2;
    }
}

//------------------------------------------------------------------------------

void setVirtualScreenWidth()
{   
    viscwi = text[curt].viscwi;

    if(prefs.sbotex && viscwi==SCREEN_WIDTH)
        viscwi -= verScroll.sizex;    
}

//------------------------------------------------------------------------------

void buttonFile(buttonType* b)
{
    showDir(textpath, "text", TEXTDIR_VIEW);                  
}

//------------------------------------------------------------------------------

void buttonFileNum(buttonType* b)
{
uint16 si, i1, i2;    
    
    text[curt].fivipo = tsfipo;                 // save current text position

    si = b->subix;                              // handle buttons
    i1 = (si + 1) % 3 + b->index - si;
    i2 = (si + 2) % 3 + b->index - si; 
    button[i1].status &= ~CHECKED;
    drawButton(&button[i1], REFRESH);
    button[i2].status &= ~CHECKED;
    drawButton(&button[i2], REFRESH);
    b->status |= CHECKED;
    drawButton(b, REFRESH);
    
    curt = si + 1;                              // set current text 1...3
    prefs.curt = curt;
    sy = 0;

    sprintf(button[BUT_SCRWIDTH].text, "%d", text[curt].viscwi);    // refresh buttons and texts    
    drawButton(&button[BUT_SCRWIDTH], REFRESH);                
    button[BUT_WORDWRAP].status = text[curt].wordwrap ? button[BUT_WORDWRAP].status | CHECKED : button[BUT_WORDWRAP].status & ~CHECKED;
    button[BUT_LINECOMB].status = text[curt].linecomb ? button[BUT_LINECOMB].status | CHECKED : button[BUT_LINECOMB].status & ~CHECKED;
    button[BUT_FORCEMONO].status = text[curt].forcemono ? button[BUT_FORCEMONO].status | CHECKED : button[BUT_FORCEMONO].status & ~CHECKED;
    drawButton(&button[BUT_WORDWRAP], REFRESH);                   
    drawButton(&button[BUT_LINECOMB], REFRESH);                   
    drawButton(&button[BUT_FORCEMONO], REFRESH);                   

    if(loadText(text[curt].name, &text[curt].fivipo, TRUE))
        status("could not open file");
    else
        status("OK %s", text[curt].name);    // draw status
 
    setVirtualScreenWidth();
    floatText(text[curt].fivipo, TRUE, FALSE, TRUE);    // seek, update sx, do not invert line, update scrollbar
    flip = TRUE;                                // force refresh of front screen    
}

//------------------------------------------------------------------------------

void buttonScreenWidth(buttonType* b)
{
static int16 a=100;
int16 v;
char s='';

    if(b->subix == 1){                          // middle button
        a = a==100 ? 500 : 100;
    }
    else{
        v = text[curt].viscwi;
        
        if(b->subix == 0){                      // minus button
            v -= a;
            if(v < SCREEN_WIDTH) v = SCREEN_WIDTH;
            s = '-';
        }
        else if(b->subix == 2){                 // plus button
            v += a;
            if(v % 100) v -= v % 100;
            if(v > MAXSCREENWIDTH) v = MAXSCREENWIDTH;
            s = '+';
        }

        sprintf(button[BUT_SCRWIDTH].text, "%d", v);    // store in button text
        drawButton(&button[BUT_SCRWIDTH], REFRESH);     // draw middle button
        text[curt].viscwi = v;                          // store in prefs
        
        setVirtualScreenWidth();
        sx = 0; sy = 0;
        floatText(tsfipo, TRUE, FALSE, TRUE);           // seek, update sx, do not invert line, update scrollbar
        flip = TRUE;                                    // force refresh of front screen    
    }

    status("virtual screen width (%c%d)", s, a);
}

//------------------------------------------------------------------------------

void buttonWordwrap(buttonType* b)
{
    text[curt].wordwrap = !text[curt].wordwrap;                    
    button[BUT_WORDWRAP].status = text[curt].wordwrap ? button[BUT_WORDWRAP].status | CHECKED : button[BUT_WORDWRAP].status & ~CHECKED;
    drawButton(&button[BUT_WORDWRAP], REFRESH);                   
    status("wordwrap");
    sx = 0; sy = 0;
    floatText(tsfipo, TRUE, FALSE, TRUE);       // seek, update sx, do not invert line, update scrollbar
    flip = TRUE;                                // force refresh of front screen    
}

//------------------------------------------------------------------------------

void buttonLinecomb(buttonType* b)
{
    text[curt].linecomb = !text[curt].linecomb;                    
    button[BUT_LINECOMB].status = text[curt].linecomb ? button[BUT_LINECOMB].status | CHECKED : button[BUT_LINECOMB].status & ~CHECKED;
    drawButton(&button[BUT_LINECOMB], REFRESH);                   
    status("line combine");
    sx = 0; sy = 0;
    floatText(tsfipo, TRUE, FALSE, TRUE);       // seek, update sx, do not invert line, update scrollbar
    flip = TRUE;                                // force refresh of front screen    
}

//------------------------------------------------------------------------------

void buttonForceMono(buttonType* b)
{
    text[curt].forcemono = !text[curt].forcemono;                    
    button[BUT_FORCEMONO].status = text[curt].forcemono ? button[BUT_FORCEMONO].status | CHECKED : button[BUT_FORCEMONO].status & ~CHECKED;
    drawButton(&button[BUT_FORCEMONO], REFRESH);                   
    status("not used");
}

//------------------------------------------------------------------------------

void buttonFont(buttonType* b)
{
    showDir(fontpath, "font", FONTDIR_VIEW);                  
}

//------------------------------------------------------------------------------

static void buttonFontHeight(buttonType* b)
{
static int16 a=1;
int16 h;
char s='';

    if(b->subix == 1){                          // middle button
        a = a==1 ? 10 : 1;
    }
    else{
        h = tmpFont.height;
        
        if(b->subix == 0){                      // minus button
            h -= a;
            if(h < 4) h = 4;
        }
        else if(b->subix == 2){                 // plus button
            h += a;
            if(h > 60) h = 60;
        }
        
        rendered = FALSE;                               // font height changed
        sprintf(button[BUT_FONTHEIGHT].text, "%d", h);  // store in button text
        drawButton(&button[BUT_FONTHEIGHT], REFRESH);   // draw middle button
        tmpFont.height = h;                             // store in tmpfont
    }

    status("fontheight%c (%c%d)", rendered?' ':'*', s, a);
}

//------------------------------------------------------------------------------

static void buttonRenderMode(buttonType* b)
{
int16 r;

    if(b->subix != 1){                              // not middle button
        r = tmpFont.render;
        
        if(b->subix == 0){                          // minus button
            if(--r < MONOCHROME) r = MONOCHROME;
        }                                           
        else if(b->subix == 2){                     // plus button
            if(++r > ANTIALIASE2) r = ANTIALIASE2;
        }

        sprintf(button[BUT_RENDERMODE].text, "%s", r?(r==ANTIALIASE?"AAN":"AA2"):"MON");    
        drawButton(&button[BUT_RENDERMODE], REFRESH);
    
        if(r != tmpFont.render){
            tmpFont.render = r;                     // store in tmpfont
            rendered = FALSE;                       // font changed
        }
    }

    status("rendermode%c", rendered?' ':'*');
}

//------------------------------------------------------------------------------

static void buttonHorSpc(buttonType* b)
{
static int16 a=10;
int16 h, i, ad;
char s='';

    if(b->subix == 1){                              // middle button
        a = a==1 ? 10 : 1;                          
    }                                               
    else{                                           
        h = tmpFont.hospco;                         
                                                    
        if(b->subix == 0){                          // minus button
            h -= a;                                 
            if(h < -30) h = -30;                    
        }                                           
        else if(b->subix == 2){                     // plus button
            h += a;                              
            if(h > 100) h = 100;                 
        }                                        
                                                  
        sprintf(button[BUT_HORSPC].text, "%d", h);  // store in button text
        drawButton(&button[BUT_HORSPC], REFRESH);   // draw middle button
        tmpFont.hospco = txtFont.hospco = h;        // store in tmpfont + txtfont
                                                   
        for(i=0; i<NUMGLYPHS; i++){                
            ad = txtFont.metrics[i].advnat * (100 + h);
            txtFont.metrics[i].advance = (ad + ad % 100) / 100;
        }

        tabGlyph(&txtFont);
        sx = 0; sy = 0;
        floatText(tsfipo, TRUE, FALSE, TRUE);       // seek, update sx, do not invert line, update scrollbar
        flip = TRUE;                                // force refresh of front screen    
    }

    status("horizontal spacing (%c%d)", s, a);
}

//------------------------------------------------------------------------------

static void buttonVerSpc(buttonType* b)
{
static int16 a=10;
int16 v, vs;
char s='';

    if(b->subix == 1){                              // middle button
        a = a==1 ? 10 : 1;                          
    }                                               
    else{                                           
        v = tmpFont.vespco;                         
                                                    
        if(b->subix == 0){                          // minus button
            v -= a;                                 
            if(v < -50) v = -50;                    
        }                                           
        else if(b->subix == 2){                     // plus button
            v += a;                              
            if(v > 100) v = 100;                 
        }                                        
                                                  
        sprintf(button[BUT_VERSPC].text, "%d", v);  // store in button text
        drawButton(&button[BUT_VERSPC], REFRESH);   // draw middle button
        tmpFont.vespco = txtFont.vespco = v;        // store in tmpfont + txtfont
        vs = txtFont.vspcnat * (100 + v);
        txtFont.vspace = (vs + vs % 100) / 100;

        sx = 0; sy = 0;
        floatText(tsfipo, TRUE, FALSE, TRUE);       // seek, update sx, do not invert line, update scrollbar
        flip = TRUE;                                // force refresh of front screen    
    }

    status("vertical spacing (%c%d)", s, a);
}

//------------------------------------------------------------------------------

static void buttonRender(buttonType* b)
{
    if(!cacheFont(TXTFONT, &txtFont))           // render & cache font
        rendered = TRUE;                        // rendering OK, update slider size
    
    sx = 0; sy = 0;
    floatText(tsfipo, TRUE, FALSE, TRUE);       // seek, update sx, do not invert line, update scrollbar
    flip = TRUE;                                // force refresh of front screen        
}

//------------------------------------------------------------------------------

static void buttonStatus(buttonType* b)
{
    if(++stanum > STATXTS-1)
        stanum = 0;
    
    drawButton(sbut, REFRESH);
}

//------------------------------------------------------------------------------

void buttonExit(buttonType* b)
{
    text[curt].fivipo = tsfipo;                 // save current text position
    savePrefs();
    cleanUp();
    reboot();
}

//------------------------------------------------------------------------------

/*
    void seekTextPos(int32 fivipo, bool updx)
    {
    int32 i, fx, fy, wmx, j, fxx;
    int16 ad;
    uint8 c, cc, d, lc;      
    
        fx = 0; fy = 0;                         // frame coordinates
        wmx = (prefs.viscwi * 3) / 4;          // word wrap min x
        lc = 255;
        
        for(i=0; i<=fivipo-tbfipo; i++){
            c = tbuf[i];
    
            if(c == LF){
                if(prefs.linecomb && i>1 && tbuf[i-2]>=' '){
                    continue;        
                }
                else{
                    fx = 0; fy += vspace; 
                }
    
                continue;
            }
            else if(c == CR){
                continue;
            }
    
            if(c == TAB){
                c = ' ';    
            }
            else if(c < FIRSTCHAR){
                c = DEL;
            }
    
            cc = c - FIRSTCHAR;
            ad = metrics[cc].advance;
    
            if(prefs.wordwrap){
                if(fx>wmx && lc<'!' && c>' '){
                    j = i + 1;
                    fxx = fx + ad;
                
                    while(fxx<prefs.viscwi && j<tblen && (d=tbuf[j++])>=' '){
                        fxx += metrics[d-FIRSTCHAR].advance;
                        if(d == ' ') break;
                    }
                    
                    if(fxx >= prefs.viscwi){
                        fx = 0; fy += vspace; 
                    }
                }
            }
    
            if(fx + ad >= prefs.viscwi){
                fx = 0; fy += vspace;    
            }
    
            lc = c;
            fx += ad;
        }    
    
        if(updx){
            if(fx < prefs.viscwi-SCREEN_WIDTH)
                sx = fx - ad;
            else
                sx = prefs.viscwi-SCREEN_WIDTH;
              
            if(sx < 0) sx = 0;                      
        }
    
        sy = fy + sy % vspace;
    }
*/

//------------------------------------------------------------------------------

/* 256 color palette mode example

    videoSetMode(MODE_5_2D | DISPLAY_BG2_ACTIVE); 
    videoSetModeSub(MODE_5_2D | DISPLAY_BG2_ACTIVE);
    vramSetBankA(VRAM_A_MAIN_BG_0x06000000); 
    vramSetBankC(VRAM_C_SUB_BG_0x06200000);
    
    BG2_CR = BG_BMP8_256x256 | BG_256_COLOR | BG_BMP_BASE(0);       // 256 color palette mode   
    BG2_XDX = 1<<8;
    BG2_XDY = 0;
    BG2_YDX = 0;
    BG2_YDY = 1<<8;
    BG2_CY = 0;
    BG2_CX = 0;

    SUB_BG2_CR = BG_BMP16_256x256;
    SUB_BG2_XDX = 1<<8;
    SUB_BG2_XDY = 0;
    SUB_BG2_YDX = 0;
    SUB_BG2_YDY = 1<<8;
    SUB_BG2_CY = 0;
    SUB_BG2_CX = 0;
    
    for(i=0; i<256; i++)                                            // make grey scale palette
        BG_PALETTE[i] = RGB15(i & 31, i & 31, i & 31) | BIT(15);    
    
    for(i=0; i<128*256; i++)                                        
        BG_GFX[i] = ((i >> 2) % 128 << 8) + (i >> 2) % 128;
    
    color = RGB15(0, 0, 31) | BIT(15);                              // blue

    for(i=0; i<256*256; i++)
        BG_GFX_SUB[i] = color;

//------------------------------------------------------------------------------

    // fast delete screen, bug in BIOS routine writes only quick the first 1/4 of pixels
    u16 color = RGB15(31, 31, 31) | BIT(15);        // white
    u32 colcol = color << 16 | color;
    swiFastCopy(&colcol, BG_GFX, SCREEN_WIDTH*SCREEN_HEIGHT/2 | COPY_MODE_WORD | COPY_MODE_FILL);
    
    // 4 byte wise screen fill
    u32 *p, *pe;

    u16 color = RGB15(31, 31, 31) | BIT(15);        // white
    u32 colcol = color << 16 | color;
    pe = BG_GFX + SCREEN_WIDTH * SCREEN_HEIGHT;
    
    p = BG_GFX; 
             
    while(p < pe)
        *p++ = colcol;
*/

/*
                if(prefs.linecomb && i>1){
                    d = tbuf[i-2];
                                        // CHR = character, NCH = noncharacter
                    if(d > ' '){        // CHR CR LF -> CHR SPC         
                        c = ' ';        
                    }
                    else if(d < ' '){   // NCH CR LF -> NCH CR LF
                        NEWLINE(1)
                    }                   // SPC CR LF -> SPC
                }
                else{                   // linecomb = FALSE
                    NEWLINE(1)
                }
            }
            
            if(c != ' ') continue;
*/

//    if(view  && licm==lic){
//        x1 = fx - sx; 
//        x2 = x1 + ad - 1; 
//        y1 = fy - sy;
//        y2 = y1 + vspace - 1;
//
//        if(x1 < 0) x1 = 0;
//        if(x2 >= SCREEN_WIDTH) x2 = SCREEN_WIDTH-1;
//        if(y1 < 0) y1 = 0;
//        invertRec(back, x1 , y1, x2, y2);
//    }

/* -----------------------------------------------------------------------------

    int16 nMinX=0, nMaxX=255, nMinY=0, nMaxY=192;
    
    void lineFill(int16 x1, int16 x2, int16 y, uint16 fill_color, uint16 seed_color)
    {
    int16 xL, xR;
        
        if(y<nMinY || nMaxY<y)
            return;
            
        for(xL=x1; xL>=nMinX; --xL){                        // scan left
            if(getPixel(xL, y) != seed_color)
                break;
                
            setPixel(xL, y, fill_color);
        }
        
        if(xL < x1){
            lineFill(xL, x1, y-1, fill_color, seed_color);  // fill child
            lineFill(xL, x1, y+1, fill_color, seed_color);  // fill child
            ++x1;
        }
        
        for(xR=x2; xR<=nMaxX; ++xR){                        // scan right
            if(getPixel(xR, y) != seed_color)
                break;
                
            setPixel(xR, y, fill_color);
        }
        
        if(xR > x2){
            lineFill(x2, xR, y-1, fill_color, seed_color);  // fill child
            lineFill(x2, xR, y+1, fill_color, seed_color);  // fill child
            --x2;
        }
        
        for(xR=x1; xR<=x2 && xR<=nMaxX; ++xR){              // scan betweens
            if(getPixel(xR, y) == seed_color)
                setPixel(xR, y, fill_color);
            else{
                if(x1 < xR){
                    lineFill(x1, xR-1, y-1, fill_color, seed_color);    // fill child
                    lineFill(x1, xR-1, y+1, fill_color, seed_color);    // fill child
                    x1 = xR;
                }
                
                for(; xR<=x2 && xR<=nMaxX; ++xR){           // this function still works if this step is removed
                    if(getPixel(xR, y) == seed_color){      // skip over border
                        x1 = xR--;
                        break;
                    }
                }
            }
        }
    }
*/

//------------------------------------------------------------------------------

/* faulty pixelwise scrolling via scrollbar
 
    dty = verScroll.slpos - oslp;
    
    if(dty){
        if(fsize<=TBUFSIZE && abs(dty)<curfont->vspace){
            if(verScroll.slpos){
                if(dty > 0)
                    sy = (fsize-tslapo) * dty / (verScroll.sizey - oslp - verScroll.slsize);
                else if(dty < 0)
                    sy = (fsize-tsfipo) * dty / (verScroll.sizey - oslp);
            }
            else{
                sy = 0;
            }
    
            floatText(-1, FALSE, FALSE, FALSE);   // no seek, no update sx, no invert line, no update scrollbar
            flip = TRUE;
        }
*/

/*------------------------------------------------------------------------------

    void greyRec(uint16 *dbuf, int16 x1, int16 y1, int16 x2, int16 y2)
    {
    int16 xDiff;
    int16 yDiff;    
    int16 xStep = 1;
    int16 yStep;
    int32 offset; 
    int16 i, j;
    uint16 c; 
       
        if(x1 < 0) x1 = 0;
        if(x2 >= SCREEN_WIDTH) x2 = SCREEN_WIDTH-1;
        if(y1 < 0) y1 = 0;
    
        xDiff = x2 - x1;
        yDiff = y2 - y1;    
        offset = y1 * SCREEN_WIDTH + x1; 
       
        if(xDiff < 0){                     
           xDiff = -xDiff;            
           xStep = -xStep;            
        }        
    
        yStep = SCREEN_WIDTH - (xDiff + 1); 
    
        if(yDiff < 0){                            
           yDiff = -yDiff;                 
           yStep = -yStep;
        }
       
        for(j=0; j<yDiff+1; j++){
            for(i=0; i<xDiff+1; i++){                           
                c = dbuf[offset];
                c = (c&0x1f) + ((c&0x1f<<5)>>5) + ((c&0x1f<<10)>>10);
                c = c ? c/3 : 12;
                dbuf[offset] = RGB15(c, c, c) | BIT(15);
                offset += xStep;
            }
            
            offset += yStep;        
        }
    }
*/

//------------------------------------------------------------------------------

/* shows faces of a font

    void chooseFontNum(int32 num)
    {
    char name[64];
    int32 i, n;
    int16 p;
        
        if(num > 0){                                            // a font was choosen
            p = 0; n = 0;
            
            for(i=0; i<tblen; i++){                             // extract its name
                if(n == num) name[p++] = tbuf[i];
                else if(n > num) break;
                if(tbuf[i] == LF) n++;
            }
    
            name[p-1] = '\0';                                   // terminate name
            showFaces(name);                                    // show faces of choosen font
        }
        else{                                                   // -1 = empty space or 0 = headline
            setView(TEXTFILE_VIEW);                             // must be done before to set curt
    
            if(loadText(text[curt].name, &text[curt].fivipo, TRUE))
                status("could not reopen file");
          
            floatText(text[curt].fivipo, TRUE, FALSE, TRUE);    // seek, update sx, do not invert line, update scrollbar
            flip = TRUE;                                        // force refresh of front screen
            lcdSwap();                                          // control to bottom screen
        }
    }
    
    //------------------------------------------------------------------------------
    
    bool showFaces(char *name)
    {
    FT_Error error; 
    FT_Face face;
    char pana[256];
    char *buf;
    int16 i, f, num_faces;
    
        strcpy(pana, fontpath);                         // path 
        strcat(pana, name);                             // + filename
        
        error = FT_Init_FreeType(&library);
        
        if(error){
            status("could not open FreeType library");
            return error;
        }
    
        error = FT_New_Face(library, pana, 0, &face);   // open face 0 
        
        if(error == FT_Err_Unknown_File_Format){
            status("unknown font file format");
            FT_Done_FreeType(library);                  // close library
            return error;
        }
        else if(error){  
            status("cannot open font file");
            FT_Done_FreeType(library);                  // close library
            return error;
        }
        
        num_faces = face->num_faces;                    // get number of faces in font
        FT_Done_Face(face);                             // close face
    
        strcpy(tmpFont.name, name);                     // font valid, copy new fontname
        buf = (char*)tbuf;                              // pointer to textbuffer
        sprintf(buf, "Load Face - glyphs [sizes] - (exit):\n");
        f = 0;                                          // start with face 0
    
        while(tblen<TBUFSIZE-100 && f<num_faces){
            if(!(error=FT_New_Face(library, pana, f, &face))){
                sprintf(buf+strlen(buf), "%d. %d", f+1, (int32)face->num_glyphs);
                
                if(face->num_fixed_sizes){
                    strcat(buf, " [");
        
                    for(i=0; i<face->num_fixed_sizes; i++)
                        sprintf(buf+strlen(buf), "%d ", (int32)FT_CEIL(face->available_sizes[i].y_ppem));
                    
                    sprintf(buf+strlen(buf)-1, "]\n");
                }
                
                FT_Done_Face(face);                     // close face
            }                                           
                                                        
            f++;                                        // next face              
        }                                               
                                                        
        FT_Done_FreeType(library);                      // close library
                                                        
        fsize = tblen = strlen(buf);                    // for scrollbars
        tbfipo = 0;                                     
        tblapo = fsize - 1;                               
                                                        
        setView(FACE_VIEW);                             // must be done before to set curt
        sx = 0; sy = 0;                                 // init text screen position
        tx = 1; ty = 1;                                 // imitate tap into headline    
        floatText(0, TRUE, TRUE, TRUE);                 // seek, update sx, invert headline, update scrollbar
        flip = TRUE;                                    // force refresh of front screen
        status("OK %.22s", name);
        return FALSE;                                   // no error
    }
    
    //------------------------------------------------------------------------------
    
    void chooseFaceNum(int32 num)
    {
        setView(TEXTFILE_VIEW);                             // must be done before to set curt
        
        if(num > 0){                                        // -1 = empty space, 0 = headline
            tmpFont.findex = num - 1;        
            rendered = FALSE;                               // font changed but not rendered
        }
    
        if(loadText(text[curt].name, &text[curt].fivipo, TRUE))
            status("could not reopen file");
        else
            status("OK%c %.22s (%d)", rendered?' ':'*', tmpFont.name, tmpFont.findex+1);
      
        floatText(text[curt].fivipo, TRUE, FALSE, TRUE);    // seek, update sx, do not invert line, update scrollbar
        flip = TRUE;                                        // force refresh of front screen
        lcdSwap();                                          // control to bottom screen
    }
*/

//------------------------------------------------------------------------------

/* LCD pixel mode on white background

    else if(bitmap->pixel_mode == FT_PIXEL_MODE_LCD){
        width = width / 3;

        while(height--){
            s = src;
            wa = width;
            xx = x;

            while(wa--){
                b = *(s++);
                g = *(s++);
                r = *(s++);
    
                if((r | g | b) && xx>-1 && xx<scrw && yy>-1)
                    *dest = RGB15((0xff-r)>>3, (0xff-g)>>3, (0xff-b)>>3) | BIT(15);
                    
                xx++;
                dest++;
            }
    
            yy++;
            src += bitmap->pitch;
            dest += SCREEN_WIDTH - width;      
        }
    }

  LCD pixel mode on color background

    else if(bitmap->pixel_mode == FT_PIXEL_MODE_LCD){
        width = width / 3;

        while(height--){
            s = src;
            wa = width;
            xx = x;

            while(wa--){
                b = *(s++);
                g = *(s++);
                r = *(s++);
    
                if((r | g | b) && xx>-1 && xx<scrw && yy>-1)
                    *dest = RGB15((0xff-r)*rr>>8, (0xff-g)*gg>>8, (0xff-b)*bb>>8) | BIT(15);
                    
                xx++;
                dest++;
            }
    
            yy++;
            src += bitmap->pitch;
            dest += SCREEN_WIDTH - width;      
        }
    }
*/
