// qdsed.h QRT 071002

#define ROOTDIR     "fat:/"
#define FONTDIR     "font/"
#define TEXTDIR     "text/"
#define PREFDIR     "font/"
#define PREFNAME    "prefs.qds"
#define SYSFONT     "sys.qds"
#define TXTFONT     "txt.qds"

#define NUMGLYPHS   256

#define BG_GFX0     ((uint16*)0x06000000)
#define BG_GFX1     ((uint16*)0x06020000)

#define TAB         0x09
#define LF          0x0a
#define CR          0x0d
#define DEL         0x7f

#define TABSIZE     4

// text buffer size
#define TBUFSIZE        16384
#define MAXSCREENWIDTH  2000
#define STATXTS         4
#define STATXTLEN       36

#define NEWLINE(x)  {li = i + x; \
                     fx = 0; \
                     fy += vspace; \
                     if(vis && fy>=sy+SCREEN_HEIGHT) break;}

enum rendermethod{
    MONOCHROME,
    ANTIALIASE,
    ANTIALIASE2
};

#define QFONTID         ('q' | ('f'<<8) | ('0'<<16) | ('1'<<24))

typedef struct{
    int16 bearX;
    int16 a_bearY;
    int16 advnat;
    int16 advance;
}metricsType;

typedef struct{
    char name[64];                  // name
    int16 height;                   // height in pixel
    int16 render;                   // render method
    int16 hospco;                   // horicontal spacing correction in percent
    int16 vespco;                   // vertical spacing correction in percent
    int16 vspcnat;                  // vertival space natural (calculated)
    int16 vspace;                   // vertical space in pixel (calculated)
    int16 nethei;                   // netto height (calculated)            
    metricsType metrics[NUMGLYPHS]; // metrics array
    FT_Glyph glyph[NUMGLYPHS];      // glyph pointer array
}qfontType;

static qfontType sysFont;
static qfontType txtFont;

static qfontType tmpFont =
{
    "arial.ttf",
    12,
    ANTIALIASE,
    0, 0, 0, 0
};

enum viewstyle{
    TEXTFILE_VIEW,
    TEXTDIR_VIEW,
    FONTDIR_VIEW
};

#define getTimer(timer) (TIMER_DATA(timer))

//-------------------------------------------------------------------------------

typedef struct{
    char  name[64];             // text name, max filesize 2 GB
    int32 fivipo;               // text first visual position
    int16 viscwi;               // virtual screen width min - 2000
    bool  wordwrap;             // wordwrapping
    bool  linecomb;             // line combining
    bool  forcemono;            // force monofont
}textType;

typedef struct{
    uint16 version;             // pref version
    bool sbotex;                // scrollbars on text
    int16 curt;                 // current text 0...3
    textType text[4];           // 3 memorized texts + text[0] for list view
}prefsType;

#define PREFVERSION 10005

static prefsType defaultPrefs =
{
    PREFVERSION,
    FALSE,
    1,

    {{"",                       // text[0] = list view  
      0,
      256,
      FALSE,
      FALSE,
      FALSE},
     
     {"doc.txt",
      0,
      500,
      TRUE,
      FALSE,
      FALSE},
      
     {"",
      0,
      500,
      TRUE,
      FALSE,
      FALSE},
     
     {"",
      0,
      500,
      TRUE,
      FALSE,
      FALSE}}
};

//-------------------------------------------------------------------------------

#define INVCOL  RGB15(17, 17, 17)

typedef struct{
    char text[8];
    int16 posx, posy;
    int16 sizex, sizey;
    int16 status;
    uint16 color1;          // unselected
    uint16 color2;          // selected
    uint16 index;
    uint16 subix;
    void *function;    
}buttonType;

static void drawButton(buttonType*, uint16);
static void buttonFile(buttonType*);
static void buttonFileNum(buttonType*);
static void buttonScreenWidth(buttonType*);
static void buttonWordwrap(buttonType*);
static void buttonLinecomb(buttonType*);
static void buttonForceMono(buttonType*);
static void buttonFont(buttonType*);
static void buttonFontHeight(buttonType*);
static void buttonRenderMode(buttonType*);
static void buttonHorSpc(buttonType*);
static void buttonVerSpc(buttonType*);
static void buttonRender(buttonType*);
static void buttonStatus(buttonType*);
static void buttonExit(buttonType*);

enum buttondraw{
    REFRESH     = 0x0000,
    INIT        = 0x0001,
    BUTDOWN     = 0x0002,
    BUTUP       = 0x0004
};

enum buttonstatus{
    LINETOP     = 0x0001,
    LINERIGHT   = 0x0002,
    LINEBOTTOM  = 0x0004,
    LINELEFT    = 0x0008,
    LINETOBO    = 0x0005,    
    LINERIBO    = 0x0006,
    LINEALL     = 0x000f,
                
    SQUARE      = 0x0000,
    TRILEFT     = 0x0010,
    TRIRIGHT    = 0x0020,
                
    PRESSED     = 0x0100,
    CHECKED     = 0x0200,
    
    CHECKBUT    = 0x2000,
    STATBUT     = 0x4000,
    LASTBUT     = 0x8000
};

#define BUT_FILENUM     1
#define BUT_SCRWIDTH    5
#define BUT_WORDWRAP    7
#define BUT_LINECOMB    8
#define BUT_FORCEMONO   9
#define BUT_FONTHEIGHT 12
#define BUT_RENDERMODE 15
#define BUT_HORSPC     18
#define BUT_VERSPC     21

buttonType button[] = {
    {"File",                // file --------------------------------------------
     0, 0,
     37, 16,
     SQUARE|LINEALL,
     RGB15(15, 25, 30),
     RGB15(0, 0, 0),
     0,
     0,
     *buttonFile},

    {"1",                   // file 1
     38, 0,
     34, 16,
     SQUARE|LINEALL|CHECKBUT,
     RGB15(15, 25, 30),
     RGB15(30, 19, 20),
     1,
     0,
     *buttonFileNum},
    
    {"2",                   // file 2
     73, 0,
     33, 16,
     SQUARE|LINEALL|CHECKBUT,
     RGB15(15, 25, 30),
     RGB15(30, 19, 20),
     2,
     1,
     *buttonFileNum},
    
    {"3",                   // file 3
     107, 0,
     34, 16,
     SQUARE|LINEALL|CHECKBUT,
     RGB15(15, 25, 30),
     RGB15(30, 19, 20),
     3,
     2,
     *buttonFileNum},

    {"",                    // screen width minus ------------------------------
     142, 0,
     10, 16,
     TRILEFT,
     RGB15(30, 24, 18),
     RGB15(30, 19, 20),
     4,
     0,
     *buttonScreenWidth},

    {"500",                 // screen width
     151, 0,
     33, 16,
     SQUARE|LINETOBO,
     RGB15(31, 31, 14),
     RGB15(0, 0, 0),
     5,
     1,
     *buttonScreenWidth},

    {"",                    // screen width plus
     183, 0,
     10, 16,
     TRIRIGHT,
     RGB15(30, 24, 18),
     RGB15(0, 0, 0),
     6,
     2,
     *buttonScreenWidth},

    {"W",                   // wordwrap ----------------------------------------
     194, 0,
     16, 16,
     SQUARE|LINEALL|CHECKBUT,
     RGB15(15, 25, 30),
     RGB15(30, 19, 20),
     7,
     0,
     *buttonWordwrap},

    {"L",                   // linecombine
     211, 0,
     16, 16,
     SQUARE|LINEALL|CHECKBUT,
     RGB15(15, 25, 30),
     RGB15(30, 19, 20),
     8,
     0,
     *buttonLinecomb},

    {"-",                   // force mono
     228, 0,
     17, 16,
     SQUARE|LINEALL|CHECKBUT,
     RGB15(15, 25, 30),
     RGB15(30, 19, 20),
     9,
     0,
     *buttonForceMono},

    {"Font",                // font --------------------------------------------
     0, 17,
     37, 16,
     SQUARE|LINEALL,
     RGB15(15, 25, 30),
     RGB15(0, 0, 0),
     10,
     0,
     *buttonFont},

    {"",                    // font height minus -------------------------------
     38, 17,
     10, 16,
     TRILEFT,
     RGB15(30, 24, 18),
     RGB15(0, 0, 0),
     11,
     0,
     *buttonFontHeight},

    {"12",                  // font height
     47, 17,
     33, 16,
     SQUARE|LINETOBO,
     RGB15(31, 31, 14),
     RGB15(0, 0, 0),
     12,
     1,
     *buttonFontHeight},

    {"",                    // font height plus
     79, 17,
     10, 16,
     TRIRIGHT,
     RGB15(30, 24, 18),
     RGB15(0, 0, 0),
     13,
     2,
     *buttonFontHeight},

    {"",                    // render mode minus -------------------------------
     90, 17,
     10, 16,
     TRILEFT,
     RGB15(30, 24, 18),
     RGB15(0, 0, 0),
     14,
     0,
     *buttonRenderMode},

    {"ANT",                 // render mode
     99, 17,
     33, 16,
     SQUARE|LINETOBO,
     RGB15(31, 31, 14),
     RGB15(0, 0, 0),
     15,
     1,
     *buttonRenderMode},

    {"",                    // render mode plus
     131, 17,
     10, 16,
     TRIRIGHT,
     RGB15(30, 24, 18),
     RGB15(0, 0, 0),
     16,
     2,
     *buttonRenderMode},

    {"",                    // horizontal spacing minus ------------------------
     142, 17,
     10, 16,
     TRILEFT,
     RGB15(30, 24, 18),
     RGB15(0, 0, 0),
     17,
     0,
     *buttonHorSpc},

    {"0",                   // horizontal spacing
     151, 17,
     33, 16,
     SQUARE|LINETOBO,
     RGB15(31, 31, 14),
     RGB15(0, 0, 0),
     18,
     1,
     *buttonHorSpc},

    {"",                    // horizontal spacing plus
     183, 17,
     10, 16,
     TRIRIGHT,
     RGB15(30, 24, 18),
     RGB15(0, 0, 0),
     19,
     2,
     *buttonHorSpc},

    {"",                    // vertical spacing minus ------------------------
     194, 17,
     10, 16,
     TRILEFT,
     RGB15(30, 24, 18),
     RGB15(0, 0, 0),
     20,
     0,
     *buttonVerSpc},

    {"0",                   // vertical spacing
     203, 17,
     33, 16,
     SQUARE|LINETOBO,
     RGB15(31, 31, 14),
     RGB15(0, 0, 0),
     21,
     1,
     *buttonVerSpc},

    {"",                    // vertical spacing plus
     235, 17,
     10, 16,
     TRIRIGHT,
     RGB15(30, 24, 18),
     RGB15(0, 0, 0),
     22,
     2,
     *buttonVerSpc},

    {"Rend",                // render font -------------------------------------
     0, 34,
     37, 16,
     SQUARE|LINEALL,
     RGB15(15, 25, 30),
     RGB15(0, 0, 0),
     23,
     0,
     *buttonRender},

    {"",                    // status ------------------------------------------
     38, 34,
     207, 16,
     SQUARE|LINEALL|STATBUT,
     RGB15(23, 28, 25),
     RGB15(0, 0, 0),
     24,
     0,
     *buttonStatus},

    {"Exit",                // exit --------------------------------------------
     0, 51,
     37, 16,
     SQUARE|LINEALL|LASTBUT,
     RGB15(15, 25, 30),
     RGB15(0, 0, 0),
     23,
     0,
     *buttonExit}
};

//------------------------------------------------------------------------------

typedef struct{
    int16 posx, posy;       // position
    int16 sizex, sizey;     // size 
    int16 slpos;            // slider position
    int16 slsize;           // slider size
    uint16 colbar1;         // color bar unselected
    uint16 colbar2;         // selected
    uint16 colsli1;         // color slider unselected 
    uint16 colsli2;         // selected
}scrollbarType;

scrollbarType verScroll =
{
    246, 0,
    10, 183,
    50, 20,
    RGB15(30, 24, 18),
    RGB15(25, 19, 13),
    RGB15(30, 19, 20),
    RGB15(25, 14, 15)
};

scrollbarType horScroll =
{
    0, 182,
    247, 10,
    50, 20,
    RGB15(30, 24, 18),
    RGB15(25, 19, 13),
    RGB15(30, 19, 20),
    RGB15(25, 14, 15)
};

buttonType butScroll = 
{
    "",
    246, 182,
    10, 10,
    SQUARE|LINERIBO,
    RGB15(30, 19, 20),
    RGB15(30, 24, 18),
    0,
    0,
    NULL
};
